from intentionally_blank.formatter import Formatter


class IdentityFormatter(Formatter):
    """The identity formatter. No changes are made.
    """
    
    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        yield from lines



from .api import (
    format_from_path_to_path,
    format_from_file_to_file,
    format_from_lines_to_file,
    format_str_to_str,
    format_from_lines_to_lines,
    transform_lines,
    list_formats,
    describe_formatter,
)

__all__ = [
    "format_from_path_to_path",
    "format_from_file_to_file",
    "format_from_lines_to_file",
    "format_str_to_str",
    "format_from_lines_to_lines",
    "transform_lines",
    "list_formats",
    "describe_formatter",
]



from functools import partial

from intentionally_blank.formatter import Formatter


class VisibleWhitespaceFormatter(Formatter):
    """Display whitespace using visible characters. Fixed tab size of eight.
    """

    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        return map(partial(transliterate_whitespace, tab_size=self.tab_size), lines)


def transliterate_whitespace(text, tab_size):
    return text.translate(str.maketrans({" ": "␣", "\t": "-" * tab_size + "→", "\n": "↵\n"}))



from collections import deque


def single(iterable):
    i = iter(iterable)
    try:
        item = next(i)
    except StopIteration:
        raise ValueError("Too few items")

    try:
        next(i)
    except StopIteration:
        return item
    raise ValueError("Too many items")


def one(item):
    yield item


def alternate_with(items, alternate_item):
    """Generate a series from items, alternating with an alternate item.
    
    items[0], alternate_item, items[1], alternate_item, ... ,items[n - 1], alternate_item
    """
    for item in items:
        yield item
        yield alternate_item


def ensure_contains(items, ensured_item):
    """Yield items, followed by ensured_item, if ensured_item is not already present.
    """
    present = False
    for item in items:
        present = present and (item == ensured_item)
        yield item
        
    if not present:
        yield ensured_item
        
        
def partition_tail(items, n):
    """Lazily partition an iterable series into a head, and tail of no more than specified length.
    
    Args:
         items: An iterable series of items.
         n: The maximum number of items to be partitioned into the tail.
         
    Returns:
        A pair of iterators, head and tail. Consuming any items from the tail iterator will cause
        the entire head iterator to be consumed, so typically the head iterator should be consumed
        before consuming any items from the tail iterator.
        
    Example:
         
         head, tail = partition_tail(range(10), 3)
         for item in head:
             print(item)  # Prints all but the last three
             
         for item in tail:
             print(item)  # Prints the last three
    """
    
    p = PartitionedTail(items, n)    
    h = HeadPartitionIterator(p)
    t = TailPartitionIterator(p)
    return h, t


class PartitionedTail:
    
    def __init__(self, items, n):
        self._i = iter(items)
        self._n = n
        self._d = deque()
        self._head_iterator = None
        self._tail_iterator = None
    
    
class HeadPartitionIterator:
    
    def __init__(self, partition_tail):
        self._pt = partition_tail
        self._pt._head_iterator = self
        
    def __iter__(self):
        return self
        
    def __next__(self):
        while len(self._pt._d) < self._pt._n:
            self._pt._d.append(next(self._pt._i))
    
        assert len(self._pt._d) == self._pt._n
        incoming_item = next(self._pt._i)  # If this raises StopIteration, allow it to propagate
        outgoing_item = self._pt._d.popleft()
        self._pt._d.append(incoming_item)
        assert len(self._pt._d) == self._pt._n
        return outgoing_item
    
    
class TailPartitionIterator:
    
    def __init__(self, partition_tail):
        self._pt = partition_tail
        self._pt._tail_iterator = self
        self._consumed_head = False
        
    def __iter__(self):
        return self
    
    def __next__(self):
        if not self._consumed_head:
            deque(self._pt._head_iterator, maxlen=0)  # Consume all items
            self._consumed_head = True
        if len(self._pt._d) == 0:
            raise StopIteration
        return self._pt._d.popleft()
    


import re

from asq import query
from more_itertools import tail

LEADING_PATTERN = r"^(\s*)(.*)$"
LEADING_REGEX = re.compile(LEADING_PATTERN)


def split_indent(line):
    m = LEADING_REGEX.match(line)
    assert m is not None
    indent, text = m.groups()
    return indent, text


def is_blank_line(index, text):
    return len(text) == 0


def strip_lines(text):
    """Remove leading and trailing blank lines."""
    is_blank = lambda line: line.isspace() or not line
    return '\n'.join(
        query(text.splitlines())
            .skip_while(is_blank)
            .reverse()
            .skip_while(is_blank)
            .reverse()
    )

def chomp(s):
    """Remove any trailing carriage-return or newline terminator."""
    if s.endswith("\r\n"): return s[:-2]
    if s.endswith("\n") or s.endswith("\r"): return s[:-1]
    return s


def split_on_newlines(s, keepends=False):
    "Split on newline only (not other characters, like str.splitlines()."
    if len(s) == 0:
        return
    lines = chomp(s).split("\n")

    for index, line in enumerate(lines):
        if keepends:
            if index == len(lines) - 1:
                # Last line
                if s.endswith("\n"):
                    end = "\n"
                else:
                    end = ""
            else:
                end = "\n"
        else:
            end = ""
            
        yield f"{line}{end}"



import sys
import textwrap
from abc import abstractmethod

from intentionally_blank.extension import ExtensionError, create_extension, Extension, list_extensions
from intentionally_blank.text import strip_lines

INTENTIONALLY_BLANK_FORMATTER = "intentionally_blank.formatter"

DEFAULT_TAB_SIZE = 4


class Formatter(Extension):

    def __init__(self, name, tab_size=None):
        super().__init__(name)
        self._tab_size = tab_size or DEFAULT_TAB_SIZE
        
    @property
    def tab_size(self):
        return self._tab_size

    @abstractmethod
    def format(self, lines):
        """Format a collection of lines.
        
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        raise NotImplementedError

    def describe(self):
        return strip_lines(textwrap.dedent(type(self).__doc__))

    def _package_name(self):
        """The name of the slide master package.
        """
        module = sys.modules[self.__module__]
        package_name = module.__package__
        return package_name


class FormatterExtensionError(ExtensionError):
    pass


def create_formatter(formatter_name, tab_size=None):
        driver = create_extension(
            kind="formatter",
            namespace=INTENTIONALLY_BLANK_FORMATTER,
            name=formatter_name,
            exception_type=FormatterExtensionError,
            tab_size=tab_size,
        )
        return driver


def formatter_names():
    return list_extensions(INTENTIONALLY_BLANK_FORMATTER)





from intentionally_blank.formatter import Formatter


class NewlineAtEofFormatter(Formatter):
    """Ensure the last line of the file has a newline terminator.
    """

    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        return map(ensure_newline_terminator, lines)
    

def ensure_newline_terminator(line):
    if line.endswith("\n"):
        return line
    return f"{line}\n"



from intentionally_blank.formatter import Formatter
from intentionally_blank.text import split_indent, is_blank_line


class EmptyBlankLineFormatter(Formatter):
    """Blank lines have no whitespace other than the terminating newline.
    """
    
    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        for line in lines:
            indent, text = split_indent(line)
            if not is_blank_line(indent, text):
                yield line
            else:
                yield "\n"



import sys

import logging
from contextlib import contextmanager

from functools import reduce

from intentionally_blank.formatter import create_formatter, formatter_names
from intentionally_blank.les_iterables import ensure_contains

logger = logging.getLogger(__name__)


def format_from_path_to_path(*, in_filepath, out_filepath=None, format_names=(), tab_size=None, in_place=False):
    """Format the contents of in_filepath and place the results in out_filepath.

    Raises:
        ValueError: If the output filepath is not provided when not processing in-place.
    """

    if (not in_place) and (out_filepath is None):
        raise ValueError("Output filepath must be provided if not processing in-place.")

    if (in_filepath == out_filepath) and (in_filepath != "-") and (not in_place):
        raise ValueError("Input and output file paths are equivalent.")

    if in_place and (out_filepath is None):
        out_filepath = in_filepath

    formatters = [
        create_formatter(format_name, tab_size)
        for format_name in ensure_contains(format_names, "identity")
    ]

    logger.debug("formatters = %s", formatters)

    in_lines = _read_lines(in_filepath)

    with open_path(out_filepath, "wt") as out_file:
        format_from_lines_to_file(in_lines, out_file, formatters, tab_size=None)


def format_from_file_to_file(in_file, out_file, format_names, tab_size=None):
    in_lines = list(in_file)
    format_from_lines_to_file(in_lines, out_file, format_names, tab_size)


def format_from_lines_to_file(in_lines, out_file, format_names, tab_size=None):
    out_file.writelines(format_from_lines_to_lines(in_lines, format_names, tab_size))


def format_str_to_str(in_text, format_names, tab_size=None):
    in_lines = in_text.splitlines(keepends=True)
    out_lines = format_from_lines_to_lines(in_lines, format_names, tab_size)
    return ''.join(out_lines)


def format_from_lines_to_lines(in_lines, format_names, tab_size=None):
    formatters = [
        create_formatter(format_name, tab_size)
        for format_name in ensure_contains(format_names, "identity")
    ]
    logger.debug("formatters = %s", formatters)
    return transform_lines(formatters, in_lines)


def transform_lines(formatters, in_lines):
    return reduce(lambda lines, formatter: formatter.format(lines), formatters, in_lines)


@contextmanager
def open_path(path, mode, encoding=None, newline=None):
    if not (("w" in mode) ^ ("r" in mode)):
        raise ValueError("Either 'r' or 'w' must be in mode")

    if not (("b" in mode) ^ ("t" in mode)):
        raise ValueError("Either 'b' or 't' must be in mode")

    if path == "-":
        f = sys.stdin if ("r" in mode) else sys.stdout
        closeable = False
    else:
        f = open(path, mode=mode, encoding=encoding, newline=newline)
        closeable = True

    try:
        yield f
    finally:
        if closeable:
            f.close()


def _read_lines(in_filepath):
    with open_path(in_filepath, "rt") as in_file:
        return list(in_file)


def list_formats(file=None):
    return list(formatter_names)

def describe_formatter(format_name, file=None):
    file = file or sys.stdout
    formatter = create_formatter(format_name)
    return formatter.describe()




from intentionally_blank.formatter import Formatter
from intentionally_blank.text import split_indent, is_blank_line


from intentionally_blank.formatter import Formatter


class RemoveTrailingFormatter(Formatter):
    """Remove any trailing whitespace except for a newline
    """

    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.

        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        return map(remove_trailing_whitespace, lines)


def remove_trailing_whitespace(line):
    """Removes trailing whitespace, but preserves the newline if present.
    """
    if line.endswith("\n"):
        return "{}\n".format(remove_trailing_whitespace(line[:-1]))
    return line.rstrip()



from abc import ABC

import stevedore
import stevedore.exception


class ExtensionError(Exception):
    pass

    
def on_load_failure(manager, entrypoint, exception):
    raise LoadFailure(manager, entrypoint, exception)
    
    
class LoadFailure(Exception):
    
    def __init__(self, manager, entrypoint, exception):
        self.manager = manager
        self.entrypoint = entrypoint
        self.__cause__ = exception
        super().__init__(f"{type(manager).__name__} failed to load entrypoint {entrypoint}")
        

def create_extension(kind, namespace, name, exception_type, *args, **kwargs):
    try:
        manager = stevedore.driver.DriverManager(
            namespace=namespace,
            name=name,
            invoke_on_load=True,
            invoke_args=args,
            invoke_kwds={**kwargs, "name": name},
            on_load_failure_callback=on_load_failure,
        )
    except stevedore.exception.NoMatches as no_matches:
        names = list_extensions(namespace)
        name_list = ", ".join(names)
        raise exception_type(
            f"No {kind} matching {name !r}. Available {kind}s: {name_list}") from no_matches
    driver = manager.driver
    return driver


def list_extensions(namespace):
    """List the names of the extensions available in a given namespace.
    """
    extensions = stevedore.ExtensionManager(
        namespace=namespace,
        invoke_on_load=False,
        on_load_failure_callback=on_load_failure,
    )
    return extensions.names()


class Extension(ABC):

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name



from intentionally_blank.formatter import Formatter
from intentionally_blank.les_iterables import partition_tail
from intentionally_blank.text import split_indent, is_blank_line


class PythonLeadingFormatter(Formatter):
    """A trailing colon introduces a new block. Blank lines have leading white space.
    """
    
    def format(self, lines):
        """
        Args:
            lines: An iterable series of strings, each with a newline terminator.
        
        Yields:
            An iterable series of strings, each with a newline terminator.
        """
        head, tail = partition_tail(lines, 1)

        # Process all but the last line
        active_indent = ""
        for line in head:
            indent, text = split_indent(line)
            if not is_blank_line(indent, text):
                yield line
                active_indent = indent
                if line.strip().endswith(":"):
                    active_indent += self.tab_size * " "
            else:
                assert line.endswith("\n")
                yield f"{active_indent}\n"
                
        # tail will contain at most one line
        for line in tail:
            indent, text = split_indent(line)
            if not is_blank_line(indent, text):
                yield line
            else:
                end = "\n" if line.endswith("\n") else ""
                if len(indent) != 0:
                    yield f"{active_indent}{end}"
                else:
                    yield end
        


import logging
import sys
from exit_codes import ExitCode

import click

from intentionally_blank import api
from intentionally_blank.formatter import formatter_names

from .version import __version__

log_levels = tuple(logging._levelToName.values())

fmt_names = formatter_names()


def _validate_optional_positive_integer(ctx, param, value):
    message = f"{param.name} needs to be a positive integer"
    if value is None:
        return value
    try:
        float_inc = float(value)
        int_inc = int(value)
    except ValueError:
        raise click.BadParameter(message)
    if float_inc != int_inc:
        raise click.BadParameter(message)
    if int_inc < 1:
        raise click.BadParameter(message)
    return int_inc


@click.group()
@click.option(
    "--verbosity",
    default="WARNING",
    help="The logging level to use.",
    type=click.Choice(log_levels, case_sensitive=True),
)
@click.version_option(version=__version__)
def cli(verbosity):
    logging_level = getattr(logging, verbosity)
    logging.basicConfig(level=logging_level)


@cli.command(name="format")
@click.option("--format", type=click.Choice(fmt_names, case_sensitive=True), multiple=True)
@click.option("--tab-size", callback=_validate_optional_positive_integer, help="Tab size in spaces")
@click.option("--in-place", "-i", is_flag=True, default=False)
# TODO: Use nargs here
@click.argument('input', type=click.Path(exists=True, dir_okay=False, allow_dash=True, resolve_path=True))
@click.argument('output', type=click.Path(dir_okay=False, allow_dash=True, resolve_path=True), required=False)
def format(input, output, format, tab_size=None, in_place=False):
    try:
        api.format_from_path_to_path(in_filepath=input, out_filepath=output, format_names=format, tab_size=tab_size, in_place=in_place)
    except ValueError as e:
        print(e, file=sys.stderr)
        sys.exit(ExitCode.NOINPUT)
    sys.exit(ExitCode.OK)


@cli.command(name="list-formats")
def list_format():
    names = api.list_formats()
    for name in names():
        print(name)
    return ExitCode.OK


@cli.command(name="describe-format")
@click.option("--format", type=click.Choice(fmt_names, case_sensitive=True), required=True)
def describe_format(format):
    description = api.describe_formatter(format_name=format)
    print(description)
    return ExitCode.OK


if __name__ == "__main__":
    cli()

